"use strict";
// Copyright (c) Mito
// Distributed under the terms of the Modified BSD License.
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const react_1 = __importDefault(require("react"));
// Import sheet and code components
const MitoSheet_1 = __importDefault(require("./MitoSheet"));
const SheetTab_1 = __importDefault(require("./SheetTab"));
const FormulaBar_1 = __importDefault(require("./FormulaBar"));
const MitoToolbar_1 = __importDefault(require("./MitoToolbar"));
const DocumentationSidebarContainer_1 = __importDefault(require("./documentation/DocumentationSidebarContainer"));
// Import modals
const ErrorModal_1 = __importDefault(require("./ErrorModal"));
const DownloadModal_1 = __importDefault(require("./DownloadModal"));
const RepeatAnalysisModal_1 = __importDefault(require("./RepeatAnalysisModal"));
const MergeModal_1 = __importDefault(require("./MergeModal"));
// Import css
require("../../css/mito.css");
const INDEX_COLUMN = "index";
var ModalEnum;
(function (ModalEnum) {
    ModalEnum["None"] = "None";
    ModalEnum["Error"] = "Error";
    ModalEnum["RepeatAnalysis"] = "RepeatAnalysis";
    ModalEnum["Download"] = "Download";
    ModalEnum["Merge"] = "Merge";
})(ModalEnum = exports.ModalEnum || (exports.ModalEnum = {}));
class Mito extends react_1.default.Component {
    constructor(props) {
        super(props);
        this.state = {
            dfNames: this.props.dfNames,
            columnSpreadsheetCodeJSONArray: this.props.columnSpreadsheetCodeJSONArray,
            sheetJSONArray: this.props.sheetJSONArray,
            selectedSheetIndex: 0,
            formulaBarValue: this.props.sheetJSONArray[0].data[0][0],
            selectedColumn: this.props.sheetJSONArray[0].columns[0].toString(),
            selectedRowIndex: 0,
            /*
                note that editingCellColumn and editingCellRowIndex should both either
                be set to these default values or they should be set to valid cell values

                If they are these default values, we are not editing a cell, and otherwise
                we are editing the cell they are set to!
            */
            editingCellColumn: "",
            editingCellRowIndex: -1,
            documentationOpen: false,
            modal: ModalEnum.None,
            errorJSON: {
                event: '',
                type: '',
                header: '',
                to_fix: ''
            }
        };
        this.cellFocused = this.cellFocused.bind(this);
        this.handleFormulaBarEdit = this.handleFormulaBarEdit.bind(this);
        this.handleFormulaBarSubmit = this.handleFormulaBarSubmit.bind(this);
        this.sendCellValueUpdate = this.sendCellValueUpdate.bind(this);
        this.setEditingMode = this.setEditingMode.bind(this);
        this.setDocumentation = this.setDocumentation.bind(this);
        this.setModal = this.setModal.bind(this);
        this.getCurrentModalComponent = this.getCurrentModalComponent.bind(this);
        this.setSelectedSheetIndex = this.setSelectedSheetIndex.bind(this);
    }
    /*
        This function is responsible for turning cell editing mode on and off
        by setting the state of editingCellColumn, and editingCellRowIndex.

        If you call this function with on===true, then the passed column and rowIndex
        should be the cell you want to start editing.

        If you call this function with on==false, then editing will be stopped, and the
        passed column and rowIndex will be focused on.
    */
    setEditingMode(on, column, rowIndex) {
        var _a, _b, _c, _d;
        if (on && this.state.editingCellRowIndex === -1) {
            /*
                This runs (and turns on editing mode) when we're not in editing mode and:
                1. The user double clicks on a cell
                2. The user presses enter on a cell.
                3. The user types any character on a cell.
            */
            this.setState({
                editingCellColumn: column,
                editingCellRowIndex: rowIndex
            });
        }
        else if (!on) {
            /*
                We turn off cell-editing mode and select the given cell.

                This handles some of the many ways cell editing can be stopped
                explicitly (e.g. ENTER), as we want only sometimes want ENTER to stop
                editing (in other cases, we want it to select a suggestion!).

                To see a list of events that stop editing, see:
                https://www.ag-grid.com/javascript-grid-cell-editing/#stop-end-editing
            */
            this.setState({
                editingCellColumn: "",
                editingCellRowIndex: -1
            });
            // We actually stop the grid from editing in this case, and set cell focus
            // as stopping editing focuses on nothing by default.
            (_b = (_a = window.gridApiMap) === null || _a === void 0 ? void 0 : _a.get(this.props.model_id)) === null || _b === void 0 ? void 0 : _b.stopEditing();
            (_d = (_c = window.gridApiMap) === null || _c === void 0 ? void 0 : _c.get(this.props.model_id)) === null || _d === void 0 ? void 0 : _d.setFocusedCell(rowIndex, column);
        }
    }
    /*
        Occurs when a cell is clicked. If we are currently editing a cell, we append the clicked
        column to the currently-edited cell. Otherwise, we select the clicked cell.
    */
    cellFocused(event) {
        var _a, _b, _c;
        /*
            We avoid cell focused throwing an error when switching sheets by making sure the
            event is defined, and just returning. Not sure why this occurs!
        */
        const column = (_a = event.column) === null || _a === void 0 ? void 0 : _a.getColId();
        if (!column)
            return;
        if (this.state.editingCellRowIndex !== -1) {
            /*
                If we're in editing mode, append the clicked column to the currently edited cell;
                We hack this by passing this info through charPress!
            */
            const selectedColumn = JSON.stringify({ selectedColumn: column });
            /*
                turn on the correct cell editor. see extensive comment in MitoCellEditor.tsx
                constructor about hacking CharPress
            */
            const editingModeParams = {
                rowIndex: this.state.editingCellRowIndex,
                colKey: this.state.editingCellColumn,
                charPress: selectedColumn
            };
            // turn the editing cell's cell editor back on!
            (_c = (_b = window.gridApiMap) === null || _b === void 0 ? void 0 : _b.get(this.props.model_id)) === null || _c === void 0 ? void 0 : _c.startEditingCell(editingModeParams);
        }
        else {
            // If we're not in editing mode, then we update the formula bar only
            // if the column is the index column, then we reset the selected cell state
            if (column === INDEX_COLUMN) {
                this.setState({
                    selectedColumn: '',
                    selectedRowIndex: 0,
                    formulaBarValue: ''
                });
                return;
            }
            // otherwise, get the cell's formula to display
            const columnIndex = this.state.sheetJSONArray[this.state.selectedSheetIndex].columns.indexOf(column);
            const rowIndex = event.rowIndex;
            const columnFormula = this.state.columnSpreadsheetCodeJSONArray[this.state.selectedSheetIndex][column];
            let formulaBarValue = '';
            if (columnFormula !== '') {
                // if the cell has a formula, display it in the formula bar
                formulaBarValue = columnFormula;
            }
            else {
                // otherwise display the value in the formula bar
                formulaBarValue = this.state.sheetJSONArray[this.state.selectedSheetIndex].data[rowIndex][columnIndex];
            }
            this.setState({
                selectedColumn: column,
                selectedRowIndex: rowIndex,
                formulaBarValue: formulaBarValue
            });
        }
    }
    handleFormulaBarEdit(e) {
        this.setState({
            formulaBarValue: e.currentTarget.value
        });
    }
    // TODO: do we want a different type of edit for a value change and a formula change or will 
    // we just detect that in the backend and apply the correct edit rules?
    handleFormulaBarSubmit(e) {
        e.preventDefault();
        this.sendCellValueUpdate(this.state.selectedColumn, this.state.formulaBarValue);
    }
    // TODO: this event should be broken out into a formula edit and a value edit
    sendCellValueUpdate(column, newValue) {
        /*
            We don't send the formula to the evaluator while in cell editing mode
            because this function gets called after the CellValueChangedEvent fires
            each time the cell editor is closed.
            
            However, the cell editor closes each time the user uses their mouse
            to reference another column - which isn't a finished update yet!
        */
        if (this.state.editingCellRowIndex === -1) {
            this.props.send({
                'event': 'edit_event',
                'type': 'cell_edit',
                'sheet_index': this.state.selectedSheetIndex,
                'id': '123',
                'timestamp': '456',
                'address': column,
                'new_formula': newValue
            });
        }
    }
    setDocumentation(documentationOpen) {
        this.setState({ documentationOpen: documentationOpen });
    }
    getCurrentModalComponent() {
        // Returns the JSX.element that is currently, open, and is used
        // in render to display this modal
        switch (this.state.modal) {
            case ModalEnum.None: return react_1.default.createElement("div", null);
            case ModalEnum.Error: return (react_1.default.createElement(ErrorModal_1.default, { errorJSON: this.state.errorJSON, setModal: this.setModal }));
            case ModalEnum.RepeatAnalysis: return (react_1.default.createElement(RepeatAnalysisModal_1.default, { setModal: this.setModal }));
            case ModalEnum.Download: return (react_1.default.createElement(DownloadModal_1.default, { setModal: this.setModal }));
            case ModalEnum.Merge: return (react_1.default.createElement(MergeModal_1.default, { setModal: this.setModal, sheetJSONArray: this.state.sheetJSONArray, dfNames: this.state.dfNames, send: this.props.send }));
        }
    }
    setModal(modal) {
        this.setState({
            'modal': modal
        });
    }
    setSelectedSheetIndex(newIndex) {
        this.setState({ selectedSheetIndex: newIndex });
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    render() {
        return (react_1.default.createElement("div", { className: "mito-container" },
            react_1.default.createElement("div", { className: "mitosheet" },
                react_1.default.createElement(MitoToolbar_1.default, { sheetJSON: this.state.sheetJSONArray[this.state.selectedSheetIndex], selectedSheetIndex: this.state.selectedSheetIndex, send: this.props.send, setDocumentation: this.setDocumentation, setModal: this.setModal, model_id: this.props.model_id }),
                react_1.default.createElement(FormulaBar_1.default, { formulaBarValue: this.state.formulaBarValue, handleFormulaBarEdit: this.handleFormulaBarEdit, handleFormulaBarSubmit: this.handleFormulaBarSubmit }),
                react_1.default.createElement(MitoSheet_1.default, { sheetJSON: this.state.sheetJSONArray[this.state.selectedSheetIndex], setEditingMode: this.setEditingMode, cellFocused: this.cellFocused, model_id: this.props.model_id, sendCellValueUpdate: this.sendCellValueUpdate }),
                react_1.default.createElement("div", { className: "sheet-tab-bar" }, this.state.sheetJSONArray.map((sheetJSON, index) => {
                    // If we can't get the df name, we just call it Sheet{index}!
                    const sheetName = this.state.dfNames[index] ? this.state.dfNames[index] : `Sheet${index + 1}`;
                    return (react_1.default.createElement(SheetTab_1.default, { key: sheetName, sheetName: sheetName, sheetIndex: index, selectedSheetIndex: this.state.selectedSheetIndex, setSelectedSheetIndex: this.setSelectedSheetIndex }));
                }))),
            this.getCurrentModalComponent(),
            this.state.documentationOpen &&
                react_1.default.createElement("div", { className: "sidebar" },
                    react_1.default.createElement(DocumentationSidebarContainer_1.default, { setDocumentation: this.setDocumentation }))));
    }
}
exports.default = Mito;
//# sourceMappingURL=Mito.js.map